@@ -4,6 +4,7 @@
   This file is part of RealtimeKit.
 
   Copyright 2009 Lennart Poettering
+  Copyright 2010 Maarten Lankhorst
 
   RealtimeKit is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -72,10 +73,20 @@
         "                        <arg name=\"thread\" type=\"t\" direction=\"in\"/>\n" \
         "                        <arg name=\"priority\" type=\"u\" direction=\"in\"/>\n" \
         "                </method>\n"                                   \
+        "                <method name=\"MakeThreadRealtimeWithPID\">\n"        \
+        "                        <arg name=\"process\" type=\"t\" direction=\"in\"/>\n" \
+        "                        <arg name=\"thread\" type=\"t\" direction=\"in\"/>\n" \
+        "                        <arg name=\"priority\" type=\"u\" direction=\"in\"/>\n" \
+        "                </method>\n"                                   \
         "                <method name=\"MakeHighPriority\">\n"          \
         "                        <arg name=\"thread\" type=\"t\" direction=\"in\"/>\n" \
         "                        <arg name=\"priority\" type=\"i\" direction=\"in\"/>\n" \
         "                </method>\n"                                   \
+        "                <method name=\"MakeHighPriorityWithPID\">\n"          \
+        "                        <arg name=\"process\" type=\"t\" direction=\"in\"/>\n" \
+        "                        <arg name=\"thread\" type=\"t\" direction=\"in\"/>\n" \
+        "                        <arg name=\"priority\" type=\"i\" direction=\"in\"/>\n" \
+        "                </method>\n"                                   \
         "                <method name=\"ResetKnown\"/>\n"               \
         "                <method name=\"ResetAll\"/>\n"                 \
         "                <method name=\"Exit\"/>\n"                     \
@@ -328,7 +339,7 @@ static void free_user(struct rtkit_user *u) {
 static bool user_in_burst(struct rtkit_user *u) {
         time_t now = time(NULL);
 
-        return now < u->timestamp + actions_burst_sec;
+        return now < u->timestamp + (time_t) actions_burst_sec;
 }
 
 static bool verify_burst(struct rtkit_user *u) {
@@ -804,7 +815,7 @@ static int process_set_realtime(struct rtkit_user *u, struct process *p, struct
                 goto finish;
         }
 
-        syslog(LOG_INFO, "Sucessfully made thread %llu of process %llu (%s) owned by '%s' RT at priority %u.\n",
+        syslog(LOG_INFO, "Successfully made thread %llu of process %llu (%s) owned by '%s' RT at priority %u.\n",
                (unsigned long long) t->pid,
                (unsigned long long) p->pid,
                get_exe_name(p->pid, exe, sizeof(exe)),
@@ -868,7 +879,7 @@ static int process_set_high_priority(struct rtkit_user *u, struct process *p, st
                 goto finish;
         }
 
-        syslog(LOG_INFO, "Sucessfully made thread %llu of process %llu (%s) owned by '%s' high priority at nice level %i.\n",
+        syslog(LOG_INFO, "Successfully made thread %llu of process %llu (%s) owned by '%s' high priority at nice level %i.\n",
                (unsigned long long) t->pid,
                (unsigned long long) p->pid,
                get_exe_name(p->pid, exe, sizeof(exe)),
@@ -1064,11 +1075,12 @@ static int lookup_client(
                 struct thread **_t,
                 DBusConnection *c,
                 DBusMessage *m,
+                pid_t pid,
                 pid_t tid) {
 
         DBusError error;
         int r;
-        unsigned long pid, uid;
+        unsigned long uid;
         unsigned long long starttime;
         struct rtkit_user *u;
         struct process *p;
@@ -1083,7 +1095,8 @@ static int lookup_client(
                 goto fail;
         }
 
-        if ((pid = get_unix_process_id(c, dbus_message_get_sender(m), &error)) == (unsigned long) -1) {
+        if (pid == (pid_t) -1 &&
+            (pid = get_unix_process_id(c, dbus_message_get_sender(m), &error)) == (pid_t) -1) {
                 syslog(LOG_ERR, "get_unix_process_id() failed: %s\n", error.message);
                 r = -EIO;
                 goto fail;
@@ -1250,7 +1263,7 @@ static int verify_canary_refusal(void) {
 
         assert_se(clock_gettime(CLOCK_MONOTONIC, &now) == 0);
 
-        if (now.tv_sec < refuse_until) {
+        if (now.tv_sec < (time_t) refuse_until) {
                 syslog(LOG_WARNING, "Recovering from system lockup, not allowing further RT threads.\n");
                 return -EPERM;
         }
@@ -1292,14 +1305,16 @@ static int handle_dbus_prop_get(const char* property, DBusMessage *r) {
 static DBusHandlerResult dbus_handler(DBusConnection *c, DBusMessage *m, void *userdata) {
         DBusError error;
         DBusMessage *r = NULL;
+        int is2 = 0;
 
         dbus_error_init(&error);
 
         /* We garbage collect on every user call */
         user_gc();
 
-        if (dbus_message_is_method_call(m, "org.freedesktop.RealtimeKit1", "MakeThreadRealtime")) {
-                uint64_t thread;
+        if (dbus_message_is_method_call(m, "org.freedesktop.RealtimeKit1", "MakeThreadRealtime") ||
+            (is2 = dbus_message_is_method_call(m, "org.freedesktop.RealtimeKit1", "MakeThreadRealtimeWithPID"))) {
+                uint64_t thread, process = (uint64_t) -1;
                 uint32_t priority;
                 struct rtkit_user *u;
                 struct process *p;
@@ -1311,18 +1326,27 @@ static DBusHandlerResult dbus_handler(DBusConnection *c, DBusMessage *m, void *u
                         goto finish;
                 }
 
-                if (!dbus_message_get_args(m, &error,
-                                           DBUS_TYPE_UINT64, &thread,
-                                           DBUS_TYPE_UINT32, &priority,
-                                           DBUS_TYPE_INVALID)) {
+                if (is2)
+                        ret = dbus_message_get_args(m, &error,
+                                                    DBUS_TYPE_UINT64, &process,
+                                                    DBUS_TYPE_UINT64, &thread,
+                                                    DBUS_TYPE_UINT32, &priority,
+                                                    DBUS_TYPE_INVALID);
+                else
+                        ret = dbus_message_get_args(m, &error,
+                                                    DBUS_TYPE_UINT64, &thread,
+                                                    DBUS_TYPE_UINT32, &priority,
+                                                    DBUS_TYPE_INVALID);
 
+                if (!ret) {
                         syslog(LOG_DEBUG, "Failed to parse MakeThreadRealtime() method call: %s\n", error.message);
                         assert_se(r = dbus_message_new_error(m, error.name, error.message));
 
                         goto finish;
                 }
 
-                if ((ret = lookup_client(&u, &p, &t, c, m, (pid_t) thread)) < 0) {
+                if ((ret = lookup_client(&u, &p, &t, c, m, (pid_t)process, (pid_t) thread)) < 0) {
+                        syslog(LOG_DEBUG, "Failed to look up client: %s\n", strerror(-ret));
                         assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), strerror(-ret)));
                         goto finish;
                 }
@@ -1339,9 +1363,10 @@ static DBusHandlerResult dbus_handler(DBusConnection *c, DBusMessage *m, void *u
 
                 assert_se(r = dbus_message_new_method_return(m));
 
-        } else if (dbus_message_is_method_call(m, "org.freedesktop.RealtimeKit1", "MakeThreadHighPriority")) {
+        } else if (dbus_message_is_method_call(m, "org.freedesktop.RealtimeKit1", "MakeThreadHighPriority")
+                   || (is2 = dbus_message_is_method_call(m, "org.freedesktop.RealtimeKit1", "MakeThreadHighPriorityWithPID"))) {
 
-                uint64_t thread;
+                uint64_t thread, process = (uint64_t) -1;
                 int32_t priority;
                 struct rtkit_user *u;
                 struct process *p;
@@ -1353,18 +1378,27 @@ static DBusHandlerResult dbus_handler(DBusConnection *c, DBusMessage *m, void *u
                         goto finish;
                 }
 
-                if (!dbus_message_get_args(m, &error,
-                                           DBUS_TYPE_UINT64, &thread,
-                                           DBUS_TYPE_INT32, &priority,
-                                           DBUS_TYPE_INVALID)) {
+                if (is2)
+                        ret = dbus_message_get_args(m, &error,
+                                                    DBUS_TYPE_UINT64, &process,
+                                                    DBUS_TYPE_UINT64, &thread,
+                                                    DBUS_TYPE_INT32, &priority,
+                                                    DBUS_TYPE_INVALID);
+                else
+                        ret = dbus_message_get_args(m, &error,
+                                                    DBUS_TYPE_UINT64, &thread,
+                                                    DBUS_TYPE_INT32, &priority,
+                                                    DBUS_TYPE_INVALID);
 
+                if (!ret) {
                         syslog(LOG_DEBUG, "Failed to parse MakeThreadHighPriority() method call: %s\n", error.message);
                         assert_se(r = dbus_message_new_error(m, error.name, error.message));
 
                         goto finish;
                 }
 
-                if ((ret = lookup_client(&u, &p, &t, c, m, (pid_t) thread)) < 0) {
+                if ((ret = lookup_client(&u, &p, &t, c, m, (pid_t)process, (pid_t) thread)) < 0) {
+                        syslog(LOG_DEBUG, "Failed to look up client: %s\n", strerror(-ret));
                         assert_se(r = dbus_message_new_error_printf(m, translate_error_forward(ret), strerror(-ret)));
                         goto finish;
                 }
@@ -1719,7 +1753,7 @@ static int drop_privileges(void) {
                 }
                 proc = "/";
 
-                syslog(LOG_DEBUG, "Sucessfully called chroot.\n");
+                syslog(LOG_DEBUG, "Successfully called chroot.\n");
         }
 
         if (do_drop_privileges) {
@@ -1772,7 +1806,7 @@ static int drop_privileges(void) {
                 setenv("LOGNAME", username, 1);
                 setenv("HOME", get_proc_path(), 1);
 
-                syslog(LOG_DEBUG, "Sucessfully dropped privileges.\n");
+                syslog(LOG_DEBUG, "Successfully dropped privileges.\n");
         }
 
         return 0;
@@ -1825,7 +1859,7 @@ static int set_resource_limits(void) {
                 }
         }
 
-        syslog(LOG_DEBUG, "Sucessfully limited resources.\n");
+        syslog(LOG_DEBUG, "Successfully limited resources.\n");
 
         return 0;
 }
@@ -2252,6 +2286,7 @@ int main(int argc, char *argv[]) {
         DBusConnection *bus = NULL;
         int ret = 1;
         struct rtkit_user *u;
+        unsigned long slack_ns;
 
         if (parse_command_line(argc, argv, &ret) <= 0)
                 goto finish;
@@ -2265,6 +2300,12 @@ int main(int argc, char *argv[]) {
                 LOG_NDELAY|LOG_PID|(log_stderr ? LOG_PERROR : 0),
                 LOG_DAEMON);
 
+        /* Raise our timer slack, we don't really need to be woken up
+         * on time. */
+        slack_ns = (((unsigned long) canary_watchdog_msec - (unsigned long) canary_cheep_msec) / 2UL) * 1000000UL;
+        if (prctl(PR_SET_TIMERSLACK, slack_ns) < 0)
+                syslog(LOG_WARNING, "PRT_SET_TIMERSLACK failed: %s\n", strerror(errno));
+
         self_drop_realtime(our_nice_level);
 
         if (setup_dbus(&bus) < 0)
